//
//  MOSettingsViewController.m
//  MOSettingsTableView
//
//  Created by Jan Christiansen on 10/1/12.
//  Copyright (c) 2012 monoid - Software Development and Consulting - Jan Christiansen.
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are met:
//
//  * Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//
//  * Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following
//  disclaimer in the documentation and/or other materials provided
//  with the distribution.
//
//  * Neither the name of Monoid - Development and Consulting -
//  Jan Christiansen nor the names of other
//  contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#import "NSArray+FunctionalStyle.h"
#import "NSDictionary+PrimitiveTypes.h"

#import "MOSettingsTableViewController.h"
#import "MOChoiceController.h"
#import "MOSwitchTableViewCell.h"
#import "MOTextTableViewCell.h"
#import "MOChoiceTableViewCell.h"
#import "MOSettingsGroupSpecification.h"
#import "MOSettingSpecification.h"
#import "MOSwitchSettingSpecification.h"
#import "MOTextSettingSpecification.h"
#import "MOChoiceSettingSpecification.h"


@interface MOSettingsTableViewController ()


@property(assign, nonatomic) BOOL respondsToPresentViewController;

@property(assign, nonatomic) BOOL respondsToDismissViewController;


// data source responds to methods
@property(assign, nonatomic) BOOL dataSourceRespondsToSetting;

@property(assign, nonatomic) BOOL dataSourceRespondsToEnableSetting;


//

@property(strong, nonatomic) NSArray *specifications;


// groups

@property(assign, nonatomic) NSUInteger numberOfGroups;

@property(strong, nonatomic) NSMutableArray *headers;

@property(strong, nonatomic) NSMutableArray *settingSpecifications;


// current settings

// array of arrays of dictionaries
@property(strong, nonatomic) NSArray *settings;


@end


@implementation MOSettingsTableViewController


- (id)initWithCoder:(NSCoder *)aDecoder
  specificationFile:(NSString *)specificationFile {

    self = [super initWithCoder:aDecoder];
    if (self) {
        [self initializeWithSpecificationFile:specificationFile];
    }
    return self;
}

- (id)initWithStyle:(UITableViewStyle)style
  specificationFile:(NSString *)specificationFile {

    NSDictionary *dict = [NSDictionary dictionaryWithContentsOfFile:specificationFile];
    MOSettingsSpecification *specification = [[MOSettingsSpecification alloc] initWithDictionary:dict];
    return [self initWithStyle:style specification:specification];
}

// array of group specifications
- (id)initWithStyle:(UITableViewStyle)style
      specification:(MOSettingsSpecification *)specification {

    self = [super initWithStyle:style];
    if (self) {
        [self initializeWithSpecification:specification];
    }
    return self;
}

- (void)initializeWithSpecificationFile:(NSString *)specificationFile {

    NSDictionary *dict = [NSDictionary dictionaryWithContentsOfFile:specificationFile];
    MOSettingsSpecification *specification = [[MOSettingsSpecification alloc] initWithDictionary:dict];
    [self initializeWithSpecification:specification];
}

- (void)initializeWithSpecification:(MOSettingsSpecification *)specification {

    _respondsToPresentViewController = [self respondsToSelector:@selector(presentViewController:animated:completion:)];
    _respondsToDismissViewController = [self respondsToSelector:@selector(dismissViewControllerAnimated:completion:)];

    _doneButtonHidden = YES;

    _localizedStrings = NO;

    self.title = NSLocalizedString(specification.title, nil);

    _numberOfGroups = specification.settingsGroups.count;
    _headers = @[].mutableCopy;
    _settingSpecifications = @[].mutableCopy;

    for (MOSettingsGroupSpecification *groupSpecification in specification.settingsGroups) {
        [_headers addObject:groupSpecification.name];
        [_settingSpecifications addObject:groupSpecification.settings];
    }

    // load array of current settings from file
    [self initSettings];
}

- (void)viewDidLoad {

    [super viewDidLoad];
}

- (void)viewWillAppear:(BOOL)animated {

    [super viewWillAppear:animated];

    if (!self.doneButtonHidden) {
        NSString *buttonTitle = self.localizedStrings ? NSLocalizedString(@"SettingsDoneButton", nil) : @"Done";
        UIBarButtonItem *backButton = [[UIBarButtonItem alloc] initWithTitle:buttonTitle
                                                                       style:UIBarButtonItemStyleDone
                                                                      target:self
                                                                      action:@selector(dismiss)];
        
        self.navigationItem.rightBarButtonItem = backButton;
    }
}

- (void)dismiss {

    [self saveSettings];

    if (_respondsToDismissViewController) {
        [self dismissViewControllerAnimated:YES completion:nil];
    } else {
        [self dismissModalViewControllerAnimated:YES];
    }
}

- (void)didReceiveMemoryWarning {

    [super didReceiveMemoryWarning];
}


#pragma mark - Setting and Getting Properties

- (void)setDataSource:(id<MOSettingsDataSource>)dataSource {

    _dataSource = dataSource;
    _dataSourceRespondsToSetting = [dataSource respondsToSelector:@selector(settingForName:inGroup:)];
    _dataSourceRespondsToEnableSetting = [dataSource respondsToSelector:@selector(enableSettingForName:inGroup:)];
}


#pragma mark - Table view data source

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView {

    return self.numberOfGroups;
}

- (NSInteger)tableView:(UITableView *)tableView
 numberOfRowsInSection:(NSInteger)section {

    NSArray *sectionSettings = [self.settings objectAtIndex:section];
    return sectionSettings.count;
}

- (UITableViewCell *)tableView:(UITableView *)tableView
         cellForRowAtIndexPath:(NSIndexPath *)indexPath {

    NSArray *sectionSpecifications = [self.settingSpecifications objectAtIndex:indexPath.section];
    MOSettingSpecification *specification = [sectionSpecifications objectAtIndex:indexPath.row];
    MOSettingTableViewCell *cell;

    NSArray *sectionSettings = [self.settings objectAtIndex:indexPath.section];
    NSDictionary *setting = [sectionSettings objectAtIndex:indexPath.row];
    if ([setting integerForKey:@"type"] == MOSwitchSetting) {
        static NSString *CellIdentifier = @"SwitchCell";
        MOSwitchTableViewCell *switchCell = [tableView dequeueReusableCellWithIdentifier:CellIdentifier];
        if (!switchCell) {
            switchCell = [[MOSwitchTableViewCell alloc] initWithReuseIdentifier:CellIdentifier];
        }
        switchCell.name = self.localizedStrings ? NSLocalizedString(specification.name, nil) : specification.name;
        switchCell.tag = specification.tag;

        if (self.dataSourceRespondsToSetting) {
            NSNumber *dataSourceResult = [self.dataSource settingForName:specification.name
                                                                 inGroup:indexPath.section];
            if (dataSourceResult) {
                switchCell.enabled = dataSourceResult.boolValue;
            } else {
                switchCell.enabled = [setting boolForKey:@"setting"];
            }
        } else {
            switchCell.enabled = [setting boolForKey:@"setting"];
        }

        cell = switchCell;
    } else if ([setting integerForKey:@"type"] == MOTextSetting) {
        static NSString *CellIdentifier = @"TextCell";
        MOTextTableViewCell *textCell = [tableView dequeueReusableCellWithIdentifier:CellIdentifier];
        if (!textCell) {
            textCell = [[MOTextTableViewCell alloc] initWithReuseIdentifier:CellIdentifier];
        }
        textCell.name = self.localizedStrings ? NSLocalizedString(specification.name, nil) : specification.name;

        NSString *dataSourceResult = [self.dataSource settingForName:specification.name
                                                             inGroup:indexPath.section];
        if (self.dataSource && dataSourceResult) {
            textCell.text = dataSourceResult;
        } else {
            textCell.text = [setting objectForKey:@"setting"];
        }

        cell = textCell;
    } else if ([setting integerForKey:@"type"] == MOChoiceSetting) {
        static NSString *CellIdentifier = @"ChoiceCell";
        MOChoiceTableViewCell *choiceCell = [tableView dequeueReusableCellWithIdentifier:CellIdentifier];
        if (!choiceCell) {
            choiceCell = [[MOChoiceTableViewCell alloc] initWithReuseIdentifier:CellIdentifier];
        }
        choiceCell.name = self.localizedStrings ? NSLocalizedString(specification.name, nil) : specification.name;
        choiceCell.choices = ((MOChoiceSettingSpecification *)specification).choices;

        NSString *dataSourceResult = [self.dataSource settingForName:specification.name
                                                             inGroup:indexPath.section];
        if (self.dataSource && dataSourceResult) {
            choiceCell.choice = dataSourceResult.integerValue;
        } else {
            choiceCell.choice = [setting integerForKey:@"setting"];
        }

        cell = choiceCell;
    }

    if (self.dataSourceRespondsToEnableSetting) {
        cell.settingEnabled = [self.dataSource enableSettingForName:specification.name
                                                            inGroup:indexPath.section];
    }

    cell.indexPath = indexPath;
    cell.delegate = self;

    return cell;
}

- (NSString *)tableView:(UITableView *)tableView
titleForHeaderInSection:(NSInteger)section {

    NSString *header = [self.headers objectAtIndex:section];
    return NSLocalizedString(header, nil);
}

- (void)tableView:(UITableView *)tableView
didSelectRowAtIndexPath:(NSIndexPath *)indexPath {

    NSArray *sectionSpecifications = [self.settingSpecifications objectAtIndex:indexPath.section];
    MOSettingSpecification *specification = [sectionSpecifications objectAtIndex:indexPath.row];

    if ([[specification class] isSubclassOfClass:[MOChoiceSettingSpecification class]]) {
        NSArray *sectionSettings = [self.settings objectAtIndex:indexPath.section];
        NSDictionary *setting = [sectionSettings objectAtIndex:indexPath.row];

        MOChoiceSettingSpecification *choiceSpecification = (MOChoiceSettingSpecification *)specification;
        MOChoiceController *choiceController = [[MOChoiceController alloc] initWithChoices:choiceSpecification.choices];
        choiceController.delegate = self;
        choiceController.name = specification.name;
        choiceController.indexPath = indexPath;
        choiceController.choice = [setting integerForKey:@"setting"];
        if (!self.navigationController) {
            NSLog(@"Choice settings are only supported within a navigation controller");
        } else {
            [self.navigationController pushViewController:choiceController
                                                 animated:YES];
        }
    }
}


#pragma mark - Reload settings by asking data source

- (void)reloadSettings {

    [self.tableView reloadData];
}


#pragma mark - Setting Changed

- (void)cellChangedState:(MOSettingTableViewCell *)cell {

    NSIndexPath *cellIndexPath = [self.tableView indexPathForCell:cell];
    NSArray *sectionSettings = [self.settings objectAtIndex:cellIndexPath.section];
    NSMutableDictionary *setting = [sectionSettings objectAtIndex:cellIndexPath.row];

    if ([[cell class] isSubclassOfClass:[MOTextTableViewCell class]]) {
        MOTextTableViewCell *textCell = (MOTextTableViewCell *)cell;
        [setting setObject:textCell.text forKey:@"setting"];
    } else if ([[cell class] isSubclassOfClass:[MOSwitchTableViewCell class]]) {
        MOSwitchTableViewCell *switchCell = (MOSwitchTableViewCell *)cell;
        [setting setObject:@(switchCell.enabled) forKey:@"setting"];
    }

    NSArray *sectionSpecifications = [self.settingSpecifications objectAtIndex:cell.indexPath.section];
    MOSettingSpecification *spec = [sectionSpecifications objectAtIndex:cell.indexPath.row];
    [self.delegate settingValueChangedWithName:spec.name
                                       inGroup:cell.indexPath.section];
}


#pragma mark - Save and Load Settings

- (NSString *)filePath {

    NSURL *documentsDir = [[[NSFileManager defaultManager] URLsForDirectory:NSDocumentDirectory
                                                                  inDomains:NSUserDomainMask] lastObject];
    NSString *fileName = [self.title stringByAppendingString:@".plist"];
    return [documentsDir URLByAppendingPathComponent:fileName].path;
}

- (void)saveSettings {

    [self.settings writeToFile:[self filePath] atomically:YES];
}

- (void)initSettings {

    NSString *filePath = [self filePath];
    if ([[NSFileManager defaultManager] fileExistsAtPath:filePath]) {
        self.settings = [[NSArray alloc] initWithContentsOfFile:filePath];
    } else {
        NSMutableArray *initialSettings = @[].mutableCopy;
        for (NSArray *sectionSpecifications in self.settingSpecifications) {
            NSMutableArray *initialSectionSettings = @[].mutableCopy;
            for (MOSettingSpecification *specification in sectionSpecifications) {
                [initialSectionSettings addObject:[specification toDictionary].mutableCopy];
            }
            [initialSettings addObject:initialSectionSettings];
        }
        self.settings = initialSettings;
    }
}


#pragma mark - Query Settings

- (id)settingForName:(NSString *)name inGroup:(NSUInteger)group {

    NSArray *sectionSpecifications = [self.settingSpecifications objectAtIndex:group];
    NSUInteger index = [sectionSpecifications indexOfObjectPassingTest:^(MOSettingSpecification *spec, NSUInteger idx, BOOL *stop) {
        return [spec.name isEqualToString:name];
    }];

    NSArray *sectionSettings = [self.settings objectAtIndex:group];
    NSDictionary *setting = [sectionSettings objectAtIndex:index];
    return [setting objectForKey:@"setting"];
}

//- (NSDictionary *)settingsForIndexPath:(NSIndexPath *)indexPath {
//
//    NSArray *sectionSpecifications = [self.settingSpecifications objectAtIndex:indexPath.section];
//    MOSettingSpecification *specification = [sectionSpecifications objectAtIndex:indexPath.row];
//
//    NSArray *sectionSettings = [self.settings objectAtIndex:indexPath.section];
//    NSDictionary *setting = [sectionSettings objectAtIndex:indexPath.row];
//
//    NSMutableDictionary *settings = [specification toDictionary].mutableCopy;
//
//    for (NSString *key in [setting allKeys]) {
//        [settings setObject:[setting objectForKey:key] forKey:key];
//    }
//
//    return settings;
//}

//- (void)setSetting:(id)value forName:(NSString *)name inGroup:(NSUInteger)group {
//
//    NSArray *sectionSpecifications = [self.settingSpecifications objectAtIndex:group];
//    NSUInteger index = [sectionSpecifications indexOfObjectPassingTest:^(MOSettingSpecification *spec, NSUInteger idx, BOOL *stop) {
//        return [spec.name isEqualToString:name];
//    }];
//
//    NSArray *sectionSettings = [self.settings objectAtIndex:group];
//    NSMutableDictionary *setting = [sectionSettings objectAtIndex:index];
//    [setting setObject:value forKey:@"setting"];
//}

- (void)enableSetting:(BOOL)enabled
              forName:(NSString *)name
              inGroup:(NSUInteger)group {

    NSArray *sectionSpecifications = [self.settingSpecifications objectAtIndex:group];
    NSUInteger index = [sectionSpecifications indexOfObjectPassingTest:^(MOSettingSpecification *spec, NSUInteger idx, BOOL *stop) {
        return [spec.name isEqualToString:name];
    }];

    NSIndexPath *settingPath = [NSIndexPath indexPathForRow:index inSection:group];
    MOSettingTableViewCell *cell = (MOSettingTableViewCell *)[self.tableView cellForRowAtIndexPath:settingPath];
    cell.settingEnabled = enabled;
}


#pragma mark - MOChoiceControllerDelegate

- (void)choiceController:(MOChoiceController *)choiceController
           didChoseIndex:(NSUInteger)choice
            forIndexPath:(NSIndexPath *)indexPath {

    NSArray *sectionSettings = [self.settings objectAtIndex:indexPath.section];
    NSMutableDictionary *setting = [sectionSettings objectAtIndex:indexPath.row];
    [setting setObject:@(choice) forKey:@"setting"];
    [self.tableView reloadData];
}


@end
